@@ -24,19 +24,25 @@
 (require 'dash)
 (require 'dash-functional)
 (require 'help-fns)
+(require 'package)
 
 (setq text-quoting-style 'grave)
 
 (defvar functions '())
 
+(defun dash-get-package-version ()
+  "Get version of dash package."
+  (with-current-buffer (find-file-noselect "dash.el")
+    (mapconcat 'number-to-string (package-desc-version (package-buffer-info)) version-separator)))
+
 (defun example-to-string (example)
   (-let* (((actual sym expected) example)
-	  (comment
-	   (cond
-	    ((eq sym '=>) (format "=> %S" expected))
-	    ((eq sym '~>) (format "~> %S" expected))
-	    ((eq sym '!!>) (format "Error"))
-	    (t (error "Invalid test case: %S" `(,actual ,sym ,expected))))))
+      (comment
+       (cond
+        ((eq sym '=>) (format "=> %S" expected))
+        ((eq sym '~>) (format "~> %S" expected))
+        ((eq sym '!!>) (format "Error"))
+        (t (error "Invalid test case: %S" `(,actual ,sym ,expected))))))
     (--> comment
       (format "%S ;; %s" actual it)
       (replace-regexp-in-string "\\\\\\?" "?" it)
@@ -164,6 +170,11 @@ FUNCTION may reference an elisp function, alias, macro or a subr."
   (search-forward s)
   (delete-char (- (length s))))
 
+(defun goto-and-replace-all (s replacement)
+  (while (progn (goto-char (point-min)) (search-forward s nil t))
+    (delete-char (- (length s)))
+    (insert replacement)))
+
 (defun create-docs-file ()
   (let ((functions (nreverse functions)))
     (with-temp-file "./README.md"
@@ -175,6 +186,8 @@ FUNCTION may reference an elisp function, alias, macro or a subr."
       (goto-and-remove "[[ function-docs ]]")
       (insert (mapconcat 'function-to-md functions "\n"))
 
+      (goto-and-replace-all "[[ version ]]" (dash-get-package-version))
+
       (simplify-quotes))))
 
 ;;; examples-to-docs.el ends here
